(* $Id: godi_pkglist.ml 977 2012-06-03 15:39:06Z gerd $ *)

open Printf

type pkg =
    { pkg_source : string;
      pkg_name : string;
      pkg_version : string;
      pkg_comment : string;
      pkg_descr : string;
      pkg_maintainer : string;
      pkg_homepage : string;
      pkg_svn_path : string;
      pkg_svn_rev : int option;
      pkg_date : float;
      pkg_rel_by : string;
      pkg_section : string;
      pkg_distfiles : string list;
    }

let read_packages_ini filename =
  let ini = new Godi_inifiles.inifile filename in
  let legacy_dec = Netencoding.Q.decode in
  let decode ?(legacy=false) sect field s =
    try
      if s <> "" && s.[0] = '"' then
        Scanf.sscanf s "%S%!" (fun s -> s)
      else
        raise Not_found
    with
      | _ ->
          if legacy then
            legacy_dec s
          else
            s in
  let getval ?(legacy=false) sect field =
    let s = ini # getval sect field in
    decode ~legacy sect field s in
  let getaval ?(legacy=false) sect field =
    let l = ini # getaval sect field in
    List.map (decode ~legacy sect field) l in
  List.map
    (fun sect ->
       try
	 let section = getval sect "section" in
	 { pkg_source = (try getval sect "source"
			 with _ -> (* legacy case *)
			   "public_godirepo_" ^ section
			);
	   pkg_name = getval sect "name";
	   pkg_version = getval sect "version";
	   pkg_comment = getval ~legacy:true sect "comment";
	   pkg_descr = getval ~legacy:true sect "descr";
	   pkg_maintainer = getval ~legacy:true sect "maintainer";
	   pkg_homepage = getval sect "homepage";
	   pkg_svn_path = getval sect "svn_path";
	   pkg_svn_rev = (let r = getval sect "svn_rev" in
			  if r = "DELETE" then None else 
			    try
			      Some (int_of_string r)
			    with
				Failure _ ->
				  failwith (filename  ^ ": Bad svn_rev in section " ^ sect)
			 );
	   pkg_date = (let s = getval sect "date" in
		       try
			 Netdate.since_epoch (Netdate.parse s)
		       with _ ->
			 failwith (filename  ^ ": Bad date in section " ^ sect)
		      );
	   pkg_rel_by = getval sect "rel_by";
	   pkg_section = section;
	   pkg_distfiles = getaval sect "distfiles";
	 }
       with
	 | Godi_inifiles.Invalid_element name ->
	     failwith (filename  ^ ": Missing attribute " ^ name ^ 
			 " in section " ^ sect)
    )
    ini#sects
;;


let write_packages_ini_ch f l =
  (* let enc = Netencoding.Q.encode in *)
  let i = ref 0 in
  List.iter
    (fun p ->
       let sect = string_of_int !i in
       output_string f (sprintf "[%s]\n" sect);
       output_string f (sprintf "source=%S\n" p.pkg_source);
       output_string f (sprintf "name=%S\n" p.pkg_name);
       output_string f (sprintf "version=%S\n" p.pkg_version);
       output_string f (sprintf "comment=%S\n" p.pkg_comment);
       output_string f (sprintf "descr=%S\n" p.pkg_descr);
       output_string f (sprintf "maintainer=%S\n" p.pkg_maintainer);
       output_string f (sprintf "homepage=%S\n" p.pkg_homepage);
       output_string f (sprintf "svn_path=%S\n" p.pkg_svn_path);
       ( let s =
	   match p.pkg_svn_rev with
	     | None -> "DELETE"
	     | Some r -> string_of_int r in
	 output_string f (sprintf "svn_rev=%S\n" s);
       );
       output_string f (sprintf "date=%S\n" (Netdate.mk_mail_date p.pkg_date));
       output_string f (sprintf "rel_by=%S\n" p.pkg_rel_by);
       output_string f (sprintf "section=%S\n" p.pkg_section);
       List.iter
	 (fun df ->
	    output_string f (sprintf "distfiles=%S\n" df))
	 p.pkg_distfiles;
       output_string f "\n";
       incr i
    )
    l
;;


let write_packages_ini filename l =
  let ch = open_out filename in
  write_packages_ini_ch ch l;
  close_out ch


let cat_re = Str.regexp "^[a-z0-9_]+"


let write_available_new_ch f l =
  List.iter
    (fun p ->
       if p.pkg_svn_rev <> None then (
	 let cat =
	   if Str.string_match cat_re p.pkg_name 0 then
	     Str.matched_string p.pkg_name
	   else
	     "UNKNOWN"
	 in
	 Printf.fprintf f "%s %s %s\n" p.pkg_name cat p.pkg_version
       )
    )
    l;
  close_out f
;;


let write_available_new filename l =
  let ch = open_out filename in
  write_available_new_ch ch l;
  close_out ch


let merge_packages_ini ~keep_only_newest pkglist_list =
  let all = Hashtbl.create 13 in
  let newest = Hashtbl.create 17 in
  
  List.iter
    (fun pkglist ->
       List.iter
         (fun pkg ->
            let delflag = pkg.pkg_svn_rev = None in
            let all_key = (pkg.pkg_name, pkg.pkg_version, delflag) in
            if Hashtbl.mem all all_key then (
              let old_pkg = Hashtbl.find all all_key in
              if pkg.pkg_date > old_pkg.pkg_date then
                Hashtbl.replace all all_key pkg
            )
            else
              Hashtbl.add all all_key pkg;
            if Hashtbl.mem newest pkg.pkg_name then (
              let old_pkg = Hashtbl.find newest pkg.pkg_name in
              if pkg.pkg_date > old_pkg.pkg_date then
                Hashtbl.replace newest pkg.pkg_name pkg
            )
            else
              Hashtbl.add newest pkg.pkg_name pkg
         )
         pkglist
    )
    pkglist_list;

  if keep_only_newest then
    Hashtbl.fold 
      (fun _ pkg acc ->
         pkg :: acc
      )
      newest
      []
  else
    Hashtbl.fold 
      (fun (n,v,d) pkg acc ->
         if d && Hashtbl.mem all (n,v,false) then (
           (* omit a deleted package if there is newer non-deleted package
              with the same version
            *)
           let other = Hashtbl.find all (n,v,false) in
           if other.pkg_date > pkg.pkg_date then
             acc
           else
             pkg :: acc
         )
         else
           pkg :: acc
      )
      all
      []


let newest_package ~ignore_deleted pkglist name =
  let newest = ref None in
  List.iter
    (fun pkg ->
       if pkg.pkg_name = name then (
         if pkg.pkg_svn_rev <> None || not ignore_deleted then (
           match !newest with
             | Some old ->
                 if pkg.pkg_date > old.pkg_date then
                   newest := Some pkg
             | None ->
                 newest := Some pkg
         )
       )
    )
    pkglist;
  !newest


let build_tgz_name p =
  (* The p.pkg_source prefix is new in godi-tools-3 *)
  p.pkg_source ^ "--" ^ p.pkg_name ^ "-" ^ p.pkg_version ^ ".build.tgz"

let legacy_build_tgz_name p =
  p.pkg_name ^ "-" ^ p.pkg_version ^ ".build.tgz"
